'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var BezierEasing = _interopDefault(require('bezier-easing'));

var performanceNow = function () {
  var now;

  if (typeof performance !== 'undefined') {
    now = function now() {
      return performance.now();
    };
  } else {
    now = function now() {
      return +new Date();
    };
  }

  return now;
}();
/**
 * @ignore
 */

var nextFrame = function () {
  if (typeof window !== 'undefined') {
    return window.requestAnimationFrame || window.webkitRequestAnimationFrame;
  }

  return function (func) {
    return setTimeout(func, 1000 / 60);
  };
}();
var subscribers = {};
var id = 0,
    isTicking = false;

var getId = function getId() {
  return id++;
};

var add = function add(func) {
  var key = "clock_".concat(getId());
  subscribers[key] = func;

  if (!isTicking) {
    nextTick();
  }

  return key;
};

var remove = function remove(id) {
  if (subscribers[id]) {
    delete subscribers[id];
  }
};

var nextTick = function nextTick() {
  var keys = Object.keys(subscribers);

  if (keys.length === 0) {
    isTicking = false;
  } else {
    var time = performanceNow();
    keys.forEach(function (key) {
      subscribers[key](time);
    });
    nextFrame(nextTick);
    isTicking = true;
  }
};

var tick = {
  add: add,
  remove: remove
};

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _superPropBase(object, property) {
  while (!Object.prototype.hasOwnProperty.call(object, property)) {
    object = _getPrototypeOf(object);
    if (object === null) break;
  }

  return object;
}

function _get(target, property, receiver) {
  if (typeof Reflect !== "undefined" && Reflect.get) {
    _get = Reflect.get;
  } else {
    _get = function _get(target, property, receiver) {
      var base = _superPropBase(target, property);

      if (!base) return;
      var desc = Object.getOwnPropertyDescriptor(base, property);

      if (desc.get) {
        return desc.get.call(receiver);
      }

      return desc.value;
    };
  }

  return _get(target, property, receiver || target);
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

/**
 * a translation from 30-seconds-of-css/#easing-variables 
 * https: //30-seconds.github.io/30-seconds-of-css/#easing-variables
 */

var easeInQuad = BezierEasing(0.55, 0.085, 0.68, 0.53);
var easeInCubic = BezierEasing(0.55, 0.055, 0.675, 0.19);
var easeInQuart = BezierEasing(0.895, 0.03, 0.685, 0.22);
var easeInQuint = BezierEasing(0.755, 0.05, 0.855, 0.06);
var easeInExpo = BezierEasing(0.95, 0.05, 0.795, 0.035);
var easeInCircle = BezierEasing(0.6, 0.04, 0.98, 0.335);
var easeOutQuad = BezierEasing(0.25, 0.46, 0.45, 0.94);
var easeOutCubic = BezierEasing(0.215, 0.61, 0.355, 1);
var easeOutQuart = BezierEasing(0.165, 0.84, 0.44, 1);
var easeOutQuint = BezierEasing(0.23, 1, 0.32, 1);
var easeOutExpo = BezierEasing(0.19, 1, 0.22, 1);
var easeOutCircle = BezierEasing(0.075, 0.82, 0.165, 1);
var easeInOutQuad = BezierEasing(0.455, 0.03, 0.515, 0.955);
var easeInOutCubic = BezierEasing(0.645, 0.045, 0.355, 1);
var easeInOutQuart = BezierEasing(0.77, 0, 0.175, 1);
var easeInOutQuint = BezierEasing(0.86, 0, 0.07, 1);
var easeInOutExpo = BezierEasing(1, 0, 0, 1);
var easeInOutCircle = BezierEasing(0.785, 0.135, 0.15, 0.86);

var TimeFunction = /*#__PURE__*/Object.freeze({
  easeInQuad: easeInQuad,
  easeInCubic: easeInCubic,
  easeInQuart: easeInQuart,
  easeInQuint: easeInQuint,
  easeInExpo: easeInExpo,
  easeInCircle: easeInCircle,
  easeOutQuad: easeOutQuad,
  easeOutCubic: easeOutCubic,
  easeOutQuart: easeOutQuart,
  easeOutQuint: easeOutQuint,
  easeOutExpo: easeOutExpo,
  easeOutCircle: easeOutCircle,
  easeInOutQuad: easeInOutQuad,
  easeInOutCubic: easeInOutCubic,
  easeInOutQuart: easeInOutQuart,
  easeInOutQuint: easeInOutQuint,
  easeInOutExpo: easeInOutExpo,
  easeInOutCircle: easeInOutCircle
});

var linearEase = function linearEase(a) {
  return a;
};
/**
 * @class Tween - simple animation
 *    
 * @example
 * new Tween({
 *  duration: 500,
 *  onStep: (t, percent) => {
 *    console.log(percent)
 *  },
 *  onEnd: () => {
 *    console.log('tween end')
 *  }
 * })
 */


var Tween =
/*#__PURE__*/
function () {
  function Tween(options) {
    _classCallCheck(this, Tween);

    this.startTime = 0;
    this.currentTime = 0;
    this.percent = 0;
    this.pausedTime = null;
    this.clockId = null;
    this.options = {};
    this.setOptions(options);
    this.step = this.step.bind(this);
  }

  _createClass(Tween, [{
    key: "setOptions",
    value: function setOptions() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
      var duration = options.duration,
          ease = options.ease;
      this.duration = duration || 0;
      this.ease = TimeFunction[ease] || linearEase;
      this.options = Object.assign({}, this.options, options);
      return this;
    }
    /**
     * @todo add support for restart after pause
     */

  }, {
    key: "start",
    value: function start() {
      // ignore start call when already in tweening
      if (this.isTweening()) return;
      this.currentTime = performanceNow();

      if (this.pausedTime) {
        this.startTime = this.startTime + (this.currentTime - this.pausedTime);
      } else {
        this.startTime = this.currentTime;
      }

      this.clockId = tick.add(this.step);
      return this;
    }
  }, {
    key: "isTweening",
    value: function isTweening() {
      return !!this.clockId;
    }
  }, {
    key: "pause",
    value: function pause() {
      tick.remove(this.clockId);
      this.clockId = null;
      this.pausedTime = performanceNow();

      if (this.options.onPause) {
        this.options.onPause(this);
      }

      return this;
    }
  }, {
    key: "stop",
    value: function stop() {
      tick.remove(this.clockId);
      this.clockId = null;
      this.percent = 0;
      this.startTime = 0;
      this.currentTime = 0;
      this.pausedTime = null;

      if (this.options.onEnd) {
        this.options.onEnd(this);
      } // console.log('tween stop')


      return this;
    }
  }, {
    key: "step",
    value: function step(time) {
      this.currentTime = time;
      var delay = this.options.delay || 0;
      var span = time - this.startTime - delay;

      if (span <= 0) {
        //do nothing during delay time 
        return;
      }

      var leftTime = this.duration - span;
      var shouldStop = false;

      if (leftTime > 0) {
        //console.log(span / this.duration)
        this.percent = span / this.duration;
      } else {
        this.percent = 1;
        shouldStop = true;
      }

      if (this.options.onStep) {
        this.options.onStep(this, this.ease(this.percent));
      }

      if (shouldStop) {
        this.stop();
      }
    }
  }]);

  return Tween;
}();

/**
 * @class Deceleration - simple animation
 *    - an animation that mimic natrual slow down during scrolling
 * @example
 * new Deceleration({
 *  velocity: 1,
 *  onStep: (movement) => {
 *    console.log(movement)
 *  },
 *  onEnd: () => {
 *    console.log('deceleration end')
 *  }
 * })
 */

var Deceleration =
/*#__PURE__*/
function () {
  function Deceleration(options) {
    _classCallCheck(this, Deceleration);

    this.step = this.step.bind(this);
    this.setOptions(options);
  }

  _createClass(Deceleration, [{
    key: "setOptions",
    value: function setOptions() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
      var velocity = options.velocity,
          _options$resistance = options.resistance,
          resistance = _options$resistance === void 0 ? 0.001 : _options$resistance;

      if (velocity) {
        this.velocity = velocity;
      }

      this.resistance = resistance;
      this.options = Object.assign({}, this.options, options);
      return this;
    }
  }, {
    key: "start",
    value: function start() {
      this.prevTime = this.currentTime = performanceNow();
      this.sign = this.velocity > 0 ? 1 : -1;
      this.prevV = Math.abs(this.velocity);
      this.clockId = tick.add(this.step);
      return this;
    }
  }, {
    key: "stop",
    value: function stop() {
      tick.remove(this.clockId);
      this.clockId = null; //console.log('deceleration stop')

      if (this.options.onEnd) {
        this.options.onEnd();
      }

      return this;
    }
  }, {
    key: "step",
    value: function step(time) {
      this.currentTime = time;
      var span = time - this.prevTime;
      var v = this.prevV;

      if (v <= 0) {
        //console.log('will stop dec', v)
        this.stop();
      } else {
        //console.log(span)
        var displacement = v * span - this.resistance * Math.pow(span, 2) / 2; //console.log(this.sign * displacement, v)

        if (this.options.onStep) {
          this.options.onStep(this.sign * displacement);
        }

        this.prevTime = this.currentTime;
        this.prevV = this.prevV - this.resistance * span;
      }
    }
  }]);

  return Deceleration;
}();

var Queue =
/*#__PURE__*/
function () {
  function Queue() {
    _classCallCheck(this, Queue);

    _defineProperty(this, "data", []);
  }

  _createClass(Queue, [{
    key: "enqueue",
    value: function enqueue(item) {
      this.data.push(item);
      return this;
    }
  }, {
    key: "dequeue",
    value: function dequeue() {
      return this.data.shift();
    }
  }]);

  return Queue;
}();

function typeOf(obj) {
  var objName = Object.prototype.toString.call(obj);
  objName = objName.slice(8, -1);

  if (objName === 'Number' && isNaN(obj)) {
    return 'NaN';
  }

  return objName;
}
function isObject(obj) {
  return typeOf(obj) === 'Object';
}
function isArray(obj) {
  return typeOf(obj) === 'Array';
}
function isString(obj) {
  return typeOf(obj) === 'String';
}
function isNumber(obj) {
  return typeOf(obj * 1) === 'Number';
}
function isFunction(obj) {
  return typeOf(obj) === 'Function';
}
function isEmptyArray(arr) {
  // assert(isArray(arr), `isEmptyArray(arr): expect arr to be type of Array, but got ${typeOf(arr)}`)
  return isArray(arr) && arr.length === 0;
}
function isEmptyObject(obj) {
  // assert(isObject(obj), `isEmptyObject(obj): expect obj to be type of Array, but got ${typeOf(obj)}`)
  return isObject(obj) && Object.keys(obj).length === 0;
}
function isEmptyString(str) {
  // assert(isString(str), `isEmptyString(str): expect str to be type of Array, but got ${typeOf(str)}`)
  return isString(str) && str === '';
}
function isEmpty(any) {
  if (isUnset(any)) return true;
  if (isEmptyString(any)) return true;
  if (isEmptyArray(any)) return true;
  if (isEmptyObject(any)) return true;
  return false;
}
function hasValue(any) {
  return isSet(any) && any !== '';
}
/**
 * we assume Undefined or Null as unsetted value
 * @params {*} value - value to be check
 * @returns {Boolean} if typeOf value is Undefined or Null return false, else return true 
 */

function isSet(value) {
  var type = typeOf(value);
  return type !== 'Undefined' && type !== 'Null';
}
/**
 * we assume Undefined or Null as unsetted value
 * @params {*} value - value to be check
 * @returns {Boolean} if typeOf value is Undefined or Null return true, else return false
 */

function isUnset(value) {
  return !isSet(value);
}
/**
 * @desc check the format for the input email 
 * 
 * @params {String} emailAddress - a email address
 * @returns {Boolean} - true | false 
 */

var isEmail = function isEmail(emailAddress) {
  var emailRE = /^(?:[-\w])+(?:\.[-\w]+)*@(?:[-\w])+(?:(?:\.[-\w]+)+)$/;

  if (emailAddress && emailRE.test(emailAddress)) {
    return true;
  }

  return false;
};
/**
 * 
 * @param {Number|String} uint - unsigned int
 * @return (true | false)
 */

function isUInt(uint) {
  return /^\d+$/.test(uint);
}
/**
 * 
 * @param {*} obj - any type in javascript
 * @return the copy of obj
 * 
 */

function deepCopy(obj) {
  var type = typeOf(obj);
  var r;

  switch (type) {
    case 'Object':
      r = {};
      Object.keys(obj).forEach(function (key) {
        r[key] = deepCopy(obj[key]);
      });
      break;

    case 'Array':
      r = [];
      obj.forEach(function (v, k) {
        r[k] = deepCopy(v);
      });
      break;

    default:
      r = obj;
  }

  return r;
}
/**
 * @desc a way to tell if two object hold the same value recursively.
 */

function deepEqual(a, b) {
  if (a == b) return true;
  var typeA = typeOf(a);
  var typeB = typeOf(b);
  if (typeA !== typeB) return false;

  switch (typeA) {
    case 'Object':
      return deepEqualObject(a, b);

    case 'Array':
      return deepEqualArray(a, b);

    default:
      return false;
  }
}

function deepEqualObject(a, b) {
  var aKeys = Object.keys(a);

  if (aKeys.length !== Object.keys(b).length) {
    return false;
  }

  var r = true;
  aKeys.forEach(function (key) {
    if (a[key] != b[key]) {
      var ta = typeOf(a[key]);
      var tb = typeOf(b[key]);

      if (ta !== tb) {
        r = false;
      } else {
        switch (ta) {
          case 'Object':
            r = deepEqualObject(a[key], b[key]);
            break;

          case 'Array':
            r = deepEqualArray(a[key], b[key]);
            break;

          default:
            r = false;
            break;
        }
      }
    }
  });
  return r;
}

function deepEqualArray(a, b) {
  if (a.length !== b.length) {
    return false;
  }

  var r = true;
  a.forEach(function (v, i) {
    if (v != b[i]) {
      var ta = typeOf(v);
      var tb = typeOf(b[i]);

      if (ta !== tb) {
        r = false;
      } else {
        switch (ta) {
          case 'Object':
            r = deepEqualObject(v, b[i]);
            break;

          case 'Array':
            r = deepEqualArray(v, b[i]);
            break;

          default:
            r = false;
            break;
        }
      }
    }
  });
  return r;
}

function shallowEqual(a, b) {
  if (a == b) return true;
  var typeA = typeOf(a);
  var typeB = typeOf(b);
  if (typeA !== typeB) return false;

  switch (typeA) {
    case 'Object':
      return shallowEqualObject(a, b);

    case 'Array':
      return shallowEqualArray(a, b);

    default:
      return false;
  }
}

function shallowEqualObject(a, b) {
  var aKeys = Object.keys(a);

  if (aKeys.length !== Object.keys(b).length) {
    return false;
  }

  var r = true;
  aKeys.forEach(function (key) {
    if (a[key] != b[key]) {
      r = false;
    }
  });
  return r;
}

function shallowEqualArray(a, b) {
  if (a.length !== b.length) {
    return false;
  }

  var r = true;
  a.forEach(function (v, i) {
    if (v != b[i]) {
      r = false;
    }
  });
  return r;
}

function assert(condition, message) {
  var assertion = isFunction(condition) ? !!condition() : !!condition;

  if (!assertion) {
    throw new Error(message);
  }
}
/**
 * @param {Number} min - a simple int 
 * @param {Number} max - a simple int
 * @returns {Number} - a random int between min and max, maybe include min but not max
 */

function randInt(min, max) {
  min = parseInt(min);
  max = parseInt(max);
  assert(isNumber(min), "randInt(min, max): expect min to be type of Number, but got ".concat(typeOf(min)));

  if (!max) {
    max = min;
    min = 0;
  }

  return Math.floor(Math.random() * (max - min) + min);
}
/**
 * 
 * @param {Number} size - the size of the return string 
 * @param {String} sourceCode - the source characters from which to generate the random string
 * @returns {String} - a random string 
 */

function randStr() {
  var size = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 6;
  var sourceCode = arguments.length > 1 ? arguments[1] : undefined;
  var code = '0123456789abcdefghijklmnopqrstuvwxyzsABCDEFGHIJKLMNOPQRSTUVWXYZ';
  var characters = Array.from({
    length: size
  });

  if (typeOf(sourceCode) === 'String') {
    code = sourceCode;
  }

  var codeLength = code.length;
  return characters.map(function () {
    return code[randInt(codeLength)];
  }).join('');
}

var padStartProto = String.prototype.padStart;
var padEndProto = String.prototype.padEnd;
function _padStart(target, length, padString) {
  return _pad('start', target, length, padString);
}
function _padEnd(target, length, padString) {
  return _pad('start', target, length, padString);
}

function _pad(type, target, length, padString) {
  var proto = type == 'start' ? padStartProto : padEndProto;

  if (typeOf(proto) === 'Function') {
    return proto.call(target, length, padString);
  }

  target = String(target);
  var targetLength = target.length;

  if (targetLength - length >= 0) {
    return target;
  }

  padString = String(padString);
  var padLength = length - targetLength;
  var padStringLength = padString.length;
  return (type === 'start' ? '' : target) + Array.from({
    length: padLength
  }).map(function (item, index) {
    return padString[index % padStringLength];
  }).join('') + (type === 'start' ? target : '');
}

var defaultFilterFunc = function defaultFilterFunc(item) {
  return isSet(item) && !isNaN(item) && item !== false;
};
/**
 * 
 * @param {Array} array - a simple array
 * @param {Function} filterFunc - a function to check if item has value 
 * @returns {Array} - a new array without false value
 */


function compact(list, filterFunc) {
  assert(isArray(list), "compact(list): expect list to be type of Array, but got ".concat(typeOf(list)));

  if (!isFunction(filterFunc)) {
    filterFunc = defaultFilterFunc;
  }

  return list.filter(filterFunc);
}
/**
 * usage - react list render
 * 
 * @param {Array<Object>} collection - a list of plain object
 * @param {String} [prefix] - add prefix for each key
 * @returns {Array<Object>} - return a copy list of plain object who's 'key' property has been setted 
 */

function addKey(list, propName) {
  var prefix = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '';
  assert(isArray(list), "addkey(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  return list.map(function (item, key) {
    if (typeOf(item.key) !== 'Undefined') {
      return item;
    }

    var copy = deepCopy(item);
    propName = propName || 'id';
    copy.key = prefix + (typeOf(copy[propName]) !== 'Undefined' ? copy[propName] : key);
    return copy;
  });
}
function head(list) {
  var size = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 1;
  assert(isArray(list), "head(list, size): expect list to be type of Array, but got ".concat(typeOf(list)));
  var realSize = Math.min(size, list.length);
  if (realSize === 0) return '';
  if (realSize === 1) return list[0];
  return list.slice(0, realSize);
}
/**
 * 
 * @param {Array} list - a simple array
 * @param {Number} size - the size of the trancated list
 * @return {Any | Array} - the trancated list or the last item 
 */

function tail(list) {
  var size = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 1;
  assert(isArray(list), "tail(list, size): expect list to be type of Array, but got ".concat(typeOf(list)));
  var realSize = Math.min(size, list.length);
  if (realSize === 0) return '';
  if (realSize === 1) return list.slice(-1)[0];
  return list.slice(-realSize);
}
/**
 * 
 * @param {Array} list - a simple array
 * @return {Any} - any type in the list
 */

function middle(list) {
  assert(isArray(list), "middle(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  var len = list.length;
  return list[Math.floor((len + 1) / 2) - 1];
}
/**
 * 
 * @param {Array} list - a simple array
 * @returns {Int} - the index of the middle item in the list 
 */

function midIndex(list) {
  assert(isArray(list), "midIndex(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  var len = list.length;
  return Math.floor((len + 1) / 2) - 1;
}
/**
 * 
 * @param {Array} list - a simple array
 * @param {Function} func - a compare function
 * @returns {Array} - a sorted list
 */

function quickSort(list) {
  var func = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : compare;
  assert(isArray(list), "quickSort(list, func): expect list to be type of Array, but got ".concat(typeOf(list)));
  var listLen = list.length;
  if (listLen < 2) return _toConsumableArray(list);
  var leftSide = [];
  var rightSide = [];
  var headItem = head(list);

  for (var i = 1; i < listLen; i++) {
    var item = list[i];

    if (func(item, headItem)) {
      rightSide.push(item);
    } else {
      leftSide.push(item);
    }
  }

  return _toConsumableArray(quickSort(leftSide, func)).concat([headItem], _toConsumableArray(quickSort(rightSide, func)));
}

function compare(a, b) {
  return a > b;
}
/**
 * 
 * @param {Array} list - a simple array
 * @param {Number} num - the size of the largest list
 * @return {Array} - the larget list
 */


function nlargest(list) {
  var num = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 1;
  assert(isArray(list), "nlargest(list, num): expect list to be type of Array, but got ".concat(typeOf(list)));
  return quickSort(list, function (a, b) {
    return a < b;
  }).slice(0, num);
}
/**
 * 
 * @param {Array} list - a simple array
 * @param {Number} num - the size of the smallest list
 * @returns {Array} - the smallest list
 */

function nsmallest(list) {
  var num = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 1;
  assert(isArray(list), "nsmallest(list, num): expect list to be type of Array, but got ".concat(typeOf(list)));
  return quickSort(list).slice(0, num);
}
/**
 * 
 * @param {Array} list - a simple array
 * @returns {Array} - return a reordered list
 */

function shuffle(list) {
  assert(isArray(list), "shuffle(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  var len = list.length;
  var r = Array.from({
    length: len
  }).map(function (_, k) {
    return list[k];
  });
  var randIdx, midValue;

  for (var i = 0; i < len - 1; i++) {
    randIdx = randInt(i + 1, len);
    midValue = r[i];
    r[i] = r[randIdx];
    r[randIdx] = midValue;
  }

  return r;
}
/**
 * 
 * @param {Array} array - a simple array
 * @returns {Array} - return a new array without doubled values
 */

function uniq(list, equalFunc) {
  assert(isArray(list), "uniq(list, equalFunc): expect list to be type of Array, but got ".concat(typeOf(list)));

  if (isString(equalFunc)) {
    var prop = equalFunc;

    equalFunc = function equalFunc(a, b) {
      return a[prop] == b[prop];
    };
  }

  if (!equalFunc) {
    equalFunc = deepEqual;
  }

  var uniqList = [];
  list.forEach(function (item) {
    if (uniqList.some(function (uniqItem) {
      return equalFunc(item, uniqItem);
    })) {
      return;
    }

    uniqList.push(item);
  });
  return uniqList;
}
function intersection(lista, listb, equalFunc) {
  assert(isArray(lista), "intersection(lista, listb, equalFunc): expect list to be type of Array, but got ".concat(typeOf(lista)));
  assert(isArray(listb), "intersection(lista, listb, equalFunc): expect list to be type of Array, but got ".concat(typeOf(listb)));

  if (isString(equalFunc)) {
    var prop = equalFunc;

    equalFunc = function equalFunc(a, b) {
      return a[prop] == b[prop];
    };
  }

  if (!equalFunc) {
    equalFunc = deepEqual;
  }

  var intersect = [];
  lista.forEach(function (itema) {
    if (listb.some(function (itemb) {
      return equalFunc(itema, itemb);
    })) {
      intersect.push(itema);
    }
  });
  return intersect;
}
function difference(lista, listb, equalFunc) {
  assert(isArray(lista), "difference(lista, listb, equalFunc): expect list to be type of Array, but got ".concat(typeOf(lista)));
  assert(isArray(listb), "difference(lista, listb, equalFunc): expect list to be type of Array, but got ".concat(typeOf(listb)));

  if (isString(equalFunc)) {
    var prop = equalFunc;

    equalFunc = function equalFunc(a, b) {
      return a[prop] == b[prop];
    };
  }

  if (!equalFunc) {
    equalFunc = deepEqual;
  }

  var r = [];
  lista.forEach(function (itema) {
    if (!listb.some(function (itemb) {
      return equalFunc(itema, itemb);
    })) {
      r.push(itema);
    }
  });
  return r;
}
function union(lista, listb, equalFunc) {
  assert(isArray(lista), "union(lista, listb, equalFunc): expect list to be type of Array, but got ".concat(typeOf(lista)));
  assert(isArray(listb), "union(lista, listb, equalFunc): expect list to be type of Array, but got ".concat(typeOf(listb)));
  return uniq(lista.concat(listb), equalFunc);
}
function groupBy(list, groupFunc) {
  assert(isArray(list), "groupBy(list, groupFunc): expect list to be type of Array, but got ".concat(typeOf(list)));

  if (isString(groupFunc)) {
    var prop = groupFunc;

    groupFunc = function groupFunc(item) {
      return item[prop];
    };
  }

  assert(isFunction(groupFunc), "groupBy(list, groupFunc): expect groupFunc to be type of String or Function, but got ".concat(typeOf(groupFunc)));
  var r = {};
  list.forEach(function (item) {
    var key = groupFunc(item);
    r[key] = r[key] || [];
    r[key].push(item);
  });
  return r;
}
function bounds(list) {
  assert(isArray(list), "bounds(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  var x0 = Infinity,
      y0 = Infinity,
      x1 = -Infinity,
      y1 = -Infinity,
      x,
      y;
  seek(list);

  function seek(l) {
    l.forEach(function (item, index) {
      if (isString(item) || isNumber(item)) {
        index === 0 && (x0 = Math.min(x0, item));
        index === 1 && (y0 = Math.min(y0, item));
        index === 0 && (x1 = Math.max(x1, item));
        index === 1 && (y1 = Math.max(y1, item));
      }

      if (isObject(item)) {
        x = item['x'] || item['left'] || 0;
        y = item['y'] || item['top'] || 0;
        x0 = Math.min(x0, x);
        y0 = Math.min(y0, y);
        x1 = Math.max(x1, x);
        y1 = Math.max(y1, y);
      }

      if (isArray(item)) {
        seek(item);
      }
    });
  }

  return [[x0, y0], [x1, y1]];
}
function contains(list, t) {
  assert(isArray(list), "contains(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  var testFunc = t;

  if (!isFunction(t)) {
    testFunc = function testFunc(item) {
      return item === t;
    };
  }

  var index = list.findIndex(testFunc);
  return index > -1;
}

function excludeProps(obj) {
  var exclude = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  assert(isObject(obj), "excludeProps(obj, exclude): expect obj to be type of Object, but got ".concat(typeOf(obj)));

  if (isString(exclude)) {
    exclude = [exclude];
  }

  var r = {};

  var shouldExclude = function shouldExclude(name) {
    if (typeOf(exclude) === 'Function') {
      return exclude(name);
    }

    return exclude.indexOf(name) > -1;
  };

  Object.keys(obj).forEach(function (name) {
    if (!shouldExclude(name)) r[name] = obj[name];
  });
  return r;
}
function getProp(obj, nameArray) {
  var defaultValue = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '';
  assert(isObject(obj), "getProp(obj, nameArray, defaultValue): expect obj to be type of Object, but got ".concat(typeOf(obj)));
  assert(isArray(nameArray), "getProp(obj, nameArray, defaultValue): expect nameArray to be type of Array, but got ".concat(typeOf(nameArray)));
  var value = obj;

  for (var i = 0, l = nameArray.length; i < l; i++) {
    value = value[nameArray[i]];

    if (isUnset(value)) {
      break;
    }
  }

  return value || defaultValue;
}
/**
 * 
 * @param {Object} obj - a plain object
 * @returns {String} - a query string 
 */

function obj2qs(obj) {
  assert(isObject(obj), "obj2qs(obj): expect obj to be type of Object, but got ".concat(typeOf(obj)));
  var params = Object.keys(obj).map(function (k) {
    var v = obj[k];
    v = isSet(v) ? v : '';
    return "".concat(k, "=").concat(v);
  });
  return params.length ? params.join('&') : '';
}
function selectProps(obj) {
  var select = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  assert(isObject(obj), "selectProps(obj, select): expect obj to be type of Object, but got ".concat(typeOf(obj)));

  if (isString(select)) {
    select = [select];
  }

  var r = {};

  var shouldSelect = function shouldSelect(name) {
    if (typeOf(select) === 'Function') {
      return select(name);
    }

    return select.indexOf(name) > -1;
  };

  Object.keys(obj).forEach(function (name) {
    if (shouldSelect(name)) r[name] = obj[name];
  });
  return r;
}
function propCompact(obj) {
  assert(isObject(obj), "propCompact(obj): expect obj to be type of Object, but got ".concat(typeOf(obj)));
  var compactedObj = {};
  Object.keys(obj).forEach(function (key) {
    var value = obj[key];

    if (!isEmpty(value)) {
      compactedObj[key] = value;
    }
  });
  return compactedObj;
}
function extend(target) {
  assert(isObject(target), "extend(target, ...args): expect target to be type of Object, but got ".concat(typeOf(target)));
  var obj;

  for (var i = 0, l = arguments.length <= 1 ? 0 : arguments.length - 1; i < l; i++) {
    obj = i + 1 < 1 || arguments.length <= i + 1 ? undefined : arguments[i + 1];

    if (isObject(obj)) {
      Object.keys(obj).forEach(function (key) {
        target[key] = obj[key];
      });
    }
  }

  return target;
}

/**
 * @desc mimic command echo
 * @param {Any} arg - currently only accept and return the first arg
 * @returns {Any} - arg 
 */

function echo(arg) {
  return arg;
}
/**
 * @desc a empty function doing nothing when called
 */

function noop() {}
/**
 * 
 * @param {Function} func -  this is function param
 * @param {Number} wait - miniseconds of delay 
 * @param  {...any} args - arg list for the func when it called after wait time
 * @returns {Number} - relay of whate setTimeout returns
 */

function delay(func) {
  var wait = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

  for (var _len = arguments.length, args = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
    args[_key - 2] = arguments[_key];
  }

  assert(isFunction(func), "delay(func, wait, ...args): expect func to be type of Function, but got ".concat(typeOf(func)));
  assert(isNumber(wait), "delay(func, wait, ...args): expect wait to be type of Number, but got ".concat(typeOf(wait)));
  return setTimeout(function () {
    func.apply(void 0, args);
  }, wait);
}
function compose() {
  for (var _len2 = arguments.length, fns = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    fns[_key2] = arguments[_key2];
  }

  return fns.reduce(function (c, fn) {
    return function () {
      return c(fn.apply(void 0, arguments));
    };
  });
}
/**
 * 
 * @param {Function} func - any function
 * @param {Number} wait - the duration that get off multiple call
 * @param {Boolean} immediate - if immediate set true, the func will get called by the first time call and won't called during the wait time
 * @returns {Function} - a wrapper of func, multiple called wrapper during wait time, it will only call func one time.
 */

function debounce(func) {
  var wait = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  var immediate = arguments.length > 2 ? arguments[2] : undefined;
  assert(isFunction(func), "debounce(func, wait, ...args): expect func to be type of Function, but got ".concat(typeOf(func)));
  assert(isNumber(wait), "debounce(func, wait, ...args): expect wait to be type of Number, but got ".concat(typeOf(wait)));
  var timeHandle = null;

  var wrapper = function wrapper() {
    for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
      args[_key3] = arguments[_key3];
    }

    if (timeHandle) {
      clearTimeout(timeHandle);
    }

    timeHandle = setTimeout(function () {
      func.apply(void 0, args);
      timeHandle = null;
    }, wait);
  };

  var immediateWrapper = function immediateWrapper() {
    if (timeHandle) {
      return;
    }

    func.apply(void 0, arguments);
    timeHandle = setTimeout(function () {
      timeHandle = null;
    }, wait);
  };

  return immediate ? immediateWrapper : wrapper;
}

var defaultTest = function defaultTest(value) {
  return !!value;
};

function guard() {
  var test = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : defaultTest;
  var safeValue = arguments.length > 1 ? arguments[1] : undefined;
  assert(isFunction(test), "guard(test, safeValue): expect test to be type of Function, but got ".concat(typeOf(test)));
  assert(isSet(safeValue), 'guard(test, safeValue): expect safeValue to be set');
  return function (value) {
    return test(value) ? value : safeValue;
  };
}

/**
 * @see https://www.cnblogs.com/dearxinli/p/7657512.html
 * 计算字符串所占的内存字节数，默认使用UTF-8的编码方式计算，也可制定为UTF-16
 * UTF-8 是一种可变长度的 Unicode 编码格式，使用一至四个字节为每个字符编码
 * 
 * 000000 - 00007F(128个代码)      0zzzzzzz(00-7F)                             一个字节
 * 000080 - 0007FF(1920个代码)     110yyyyy(C0-DF) 10zzzzzz(80-BF)             两个字节
 * 000800 - 00D7FF 
   00E000 - 00FFFF(61440个代码)    1110xxxx(E0-EF) 10yyyyyy 10zzzzzz           三个字节
  * 010000 - 10FFFF(1048576个代码)  11110www(F0-F7) 10xxxxxx 10yyyyyy 10zzzzzz  四个字节
  * 
  * 注: Unicode在范围 D800-DFFF 中不存在任何字符
  * @see http://zh.wikipedia.org/wiki/UTF-8
  * 
  * UTF-16 大部分使用两个字节编码，编码超出 65535 的使用四个字节
  * 000000 - 00FFFF  两个字节
  * 010000 - 10FFFF  四个字节
  * 
  * @see http://zh.wikipedia.org/wiki/UTF-16
  * @param  {String} str 
  * @param  {String} charset utf-8, utf-16
  * @return {Number}
  */

function charLength(str, charset) {
  var total = 0;
  var charCode, i, len;
  charset = charset ? charset.toLowerCase() : '';

  if (charset === 'utf-16' || charset === 'utf16') {
    for (i = 0, len = str.length; i < len; i++) {
      charCode = str.charCodeAt(i);

      if (charCode <= 0xffff) {
        total += 2;
      } else {
        total += 4;
      }
    }
  } else {
    for (i = 0, len = str.length; i < len; i++) {
      charCode = str.charCodeAt(i);

      if (charCode <= 0x007f) {
        total += 1;
      } else if (charCode <= 0x07ff) {
        total += 2;
      } else if (charCode <= 0xffff) {
        total += 3;
      } else {
        total += 4;
      }
    }
  }

  return total;
}
/**
 * 
 * @param {String} str - a string that to be trim spaces at begin or end of the string.
 * @returns {String} a string without spaces at begin or end of itself.
 */

function trim(str) {
  if (isUnset(str)) {
    // if str is undefined or null, return empty string
    return '';
  }

  return (str + '').replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g, '');
}
function dashToCamel(dashStr) {
  assert(isString(dashStr), "dashToCamel(dashStr): expect dashStr to be type of String, but got ".concat(typeOf(dashStr)));

  if (dashStr.indexOf('-') === -1) {
    return dashStr;
  }

  var camelStr = dashStr;
  camelStr = trim(camelStr); // remove the dash at begin or end

  camelStr = camelStr.replace(/^-|-$/g, '');
  camelStr = camelStr.replace(/-(\w)/g, function (m, w) {
    return w.toUpperCase();
  });
  return camelStr;
}
/**
 * 
 * @param {Number|String} num - the number going to be padded
 * @param {NUmber} [size] - the length of the return string after padding
 * @returns {String} - the padded string
 */

function padZero(num) {
  var size = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 2;
  return _padStart(num, size, 0);
}
function toFixed(num) {
  var digits = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 2;
  var factor = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;
  return (num * factor).toFixed(digits);
}

var currentTime = function currentTime() {
  return +new Date();
};

var supportedFormats = ['yyyy-mm-dd hh:mm:ss', 'yyyy-mm-dd hh:mm', 'yyyy-mm-dd', 'yyyy-mm', 'hh:mm:ss', 'hh:mm'];
var supportedTargetTypes = ['seconds', 'miniseconds', 'date', 'dateString'];
/**
 * the converter of date
 * 
 */

/**
 * 
 * @param {String|Number|Date} from - input to be convert
 * @param {String} targetType - output types (seconds|miniseconds|date|dateString)
 * @param {String} [format] - supported format as follow:
 *                          'yyyy-mm-dd hh:mm:ss'
 *                          'yyyy-mm-dd hh:mm'
 *                          'yyyy-mm-dd'
 *                          'yyyy-mm'
 *                          'hh:mm:ss'
 *                          'hh:mm'
 */

function _dateConvert() {
  var from = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : currentTime();
  var targetType = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'dateString';
  var format = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 'yyyy-mm-dd hh:mm';
  var fromType = typeOf(from);

  switch (fromType) {
    case 'String':
    case 'Number':
      if (isUInt(from)) {
        return secondsConvert(ensureMiniseconds(from), targetType, format);
      }

      return DSConvert(from, targetType);

    case 'Date':
      return DConvert(from, targetType, format);

    default:
      throw new TypeError("_dateConvert - unexpected type of ".concat(fromType));
  }
}

function secondsConvert(from, targetType, format) {
  if (!(targetType === 'date' || targetType === 'dateString')) {
    throw new TypeError("unexpected targetType: ".concat(targetType));
  }

  var d = new Date();
  d.setTime(from);

  if (targetType === 'date') {
    return d;
  }

  return toDS(d, format);
}

function DSConvert(from, targetType) {
  if (!supportedTargetTypes.includes(targetType)) {
    throw new TypeError("unexpected targetType: ".concat(targetType));
  }

  if (targetType === 'dateString') {
    return from;
  }

  var reg1 = /^(\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}):(\d{2})$/;
  var reg2 = /^(\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2})$/;
  var reg3 = /^(\d{4})-(\d{2})-(\d{2})$/;
  var reg4 = /^(\d{4})-(\d{2})$/;
  var reg5 = /^(\d{2}):(\d{2}):(\d{2})$/;
  var reg6 = /^(\d{2}):(\d{2})$/;
  var today = new Date();
  var DS = String(from);
  var match, d;

  if (match = DS.match(reg1)) {
    d = new Date(match[1], match[2] - 1, match[3], match[4], match[5], match[6]);
  }

  if (match = DS.match(reg2)) {
    d = new Date(match[1], match[2] - 1, match[3], match[4], match[5]);
  }

  if (match = DS.match(reg3)) {
    d = new Date(match[1], match[2] - 1, match[3]);
  }

  if (match = DS.match(reg4)) {
    d = new Date(match[1], match[2] - 1);
  }

  if (match = DS.match(reg5)) {
    d = new Date(today.getFullYear(), today.getMonth(), today.getDate(), match[1], match[2], match[3]);
  }

  if (match = DS.match(reg6)) {
    d = new Date(today.getFullYear(), today.getMonth(), today.getDate(), match[1], match[2]);
  }

  if (!d) {
    throw new Error('unexpected input!');
  }

  if (targetType === 'seconds') {
    return Math.round(d.getTime() / 1000);
  }

  if (targetType === 'miniseconds') {
    return d.getTime();
  }

  return d;
}

function DConvert(from, targetType, format) {
  switch (targetType) {
    case 'seconds':
      return Math.round(from.getTime() / 1000);

    case 'miniseconds':
      return from.getTime();

    case 'dateString':
      return toDS(from, format);

    default:
      throw new TypeError("_dateConvert - unexpected type of ".concat(targetType));
  }
}

function ensureMiniseconds(number) {
  if ((number + '').length === 10) {
    return number * 1000;
  }

  return number;
}

function toDS(d, format) {
  if (typeOf(d) !== 'Date') {
    throw new TypeError("unexpected type: ".concat(typeOf(d)));
  }

  if (!supportedFormats.includes(String(format).toLowerCase())) {
    throw new TypeError("unexpected format! suported formats: \n".concat(supportedFormats.join('\n')));
  }

  var year,
      month,
      date,
      hours,
      minutes,
      seconds,
      dateString = format;
  year = d.getFullYear();
  month = d.getMonth() + 1;
  date = d.getDate();
  hours = d.getHours();
  minutes = d.getMinutes();
  seconds = d.getSeconds();
  dateString = dateString.replace(/yyyy/i, year);
  dateString = dateString.replace(/-mm/i, "-".concat(padZero(month)));
  dateString = dateString.replace(/dd/i, padZero(date));
  dateString = dateString.replace(/hh/i, padZero(hours));
  dateString = dateString.replace(/:mm/i, ":".concat(padZero(minutes)));
  dateString = dateString.replace(/ss/i, padZero(seconds));
  return dateString;
}

/**
 * 
 * @param {String|Number} num -  1000 | 10000000.001
 * @returns {String} 1,000 | 10,000,000.001
 */

function addComma(num, comma) {
  var numberToAddComma = parseFloat(num);

  if (isNaN(numberToAddComma)) {
    return num;
  }

  if (!isSet(comma)) {
    comma = ',';
  }

  var sign = numberToAddComma < 0 ? '-' : '';
  var base = parseInt(Math.abs(numberToAddComma)) + '';
  var decimal = (numberToAddComma + '').replace(/-?\d*/, '');
  var mod = base.length > 3 ? base.length % 3 : 0;
  return sign + (mod ? base.substr(0, mod) + comma : '') + base.substr(mod).replace(/(\d{3})(?=\d)/g, '$1' + comma) + decimal;
}
function angle2deg(angle) {
  return angle * 180 / Math.PI;
}
function deg2angle(deg) {
  return deg * Math.PI / 180;
}
/**
 * @returns {Number} - miniseconds since 00:00:00 UTC on January 1, 1970
 */

var currentTime$1 = function currentTime() {
  return +new Date();
};
/**
 * @desc - like currentTime but its seconds not miniseconds
 * @returns - seconds since 00:00:00 UTC on January 1, 1970
 */

function timestamp() {
  return Math.round(currentTime$1() / 1000);
}
/**
 * 
 * @param {Number} [startValue] - start value for generating ids
 * @returns {Function} - id will plus 1 at each call of the func
 * 
 */

function IDFactory() {
  var startValue = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;
  return function () {
    return startValue++;
  };
}
function parseQuery(queryString) {
  assert(isString(queryString), "parseQuery(queryString): expect queryString to be type of String, but got ".concat(typeOf(queryString)));
  var queryObject = {},
      str = queryString,
      lastIdx,
      params;
  lastIdx = queryString.lastIndexOf('?');

  if (lastIdx > -1) {
    str = queryString.substring(lastIdx + 1);
  }

  params = str.split('&');
  params.forEach(function (param) {
    var _param$split = param.split('='),
        _param$split2 = _slicedToArray(_param$split, 2),
        k = _param$split2[0],
        v = _param$split2[1];

    if (isSet(k)) {
      queryObject[k] = isSet(v) ? v : '';
    }
  });
  return queryObject;
}
/**
 * 
 * @param {String} left - a url or query string
 * @param {String} right - a query string
 * @returns {String} - a combined string
 */

function queryJoin(left, right) {
  var joinCharacter;
  var alreadHasQuery = /[%\w]+=/.test(left);

  if (!alreadHasQuery) {
    joinCharacter = /\?$/.test(left) ? '' : '?';
  }

  if (alreadHasQuery) {
    joinCharacter = /&$/.test(left) ? '' : '&';
  }

  return [left, joinCharacter, right].join('');
}
function sample1(list) {
  assert(isArray(list), "sample1(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  return list[randInt(list.length)];
}
function sample(list) {
  assert(isArray(list), "sample(list): expect list to be type of Array, but got ".concat(typeOf(list)));
  return shuffle(list).slice(0, randInt(list.length));
}
/**
 * 
 * @param {String} dateString 
 * @param {String} format 
 * 
 * @returns {Date} - instance of Date
 */

function strToDate(dateString, format) {
  if (!isSet(dateString)) {
    return new Date();
  }

  return _dateConvert(dateString, 'date', format);
}
/**
 * 
 * @param {Sting} dateString 
 * @param {String} format - supported formats
 * @param {Boolean} seconds - true | false
 * 
 * @returns {String} - '1543047456700'
 */

function strToTime(dateString, format, seconds) {
  if (!isSet(dateString)) {
    return seconds ? timestamp() : currentTime$1();
  }

  return _dateConvert(dateString, seconds ? 'seconds' : 'miniseconds', format);
}
/**
 * 
 * @param {String|Number} time - timestamp, maybe miniseconds
 * @param {String} [format] - supported format as follow:
 *                          'yyyy-mm-dd hh:mm:ss'
 *                          'yyyy-mm-dd hh:mm'
 *                          'yyyy-mm-dd'
 *                          'yyyy-mm'
 *                          'hh:mm:ss'
 *                          'hh:mm'
 * @returns {String} - date string 2018-11-28, 2018-11-28 14:03:23, 2000-01-01 01:01
 */

function timeToStr() {
  var time = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : timestamp();
  var format = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'yyyy-mm-dd hh:mm';
  return _dateConvert(time, 'dateString', format);
}
function reverse(obj) {
  assert(isArray(obj) || isObject(obj) || isString(obj), "reverse(min): expect obj to be type of (Array|Object|String), but got ".concat(typeOf(obj)));

  if (isArray(obj)) {
    return obj.reverse();
  }

  if (isString(obj)) {
    return obj.split('').reverse().join('');
  }

  var reverseObj = {};
  Object.keys(obj).forEach(function (key) {
    var v = obj[key];
    reverseObj[v] = key;
  });
  return reverseObj;
}
function encodeJSON(obj, replacer, space) {
  try {
    return JSON.stringify(obj, replacer, space);
  } catch (error) {
    console.log(error);
  }

  return obj;
}
function decodeJSON(str, reviver) {
  try {
    return JSON.parse(str, reviver);
  } catch (error) {
    console.log(error);
  }

  return str;
}
function safeDivide(a, b) {
  a = parseFloat(a) || 0;
  b = parseFloat(b) || 0;

  if (b === 0) {
    return 0;
  }

  return a / b;
}

var styleKeys = null;
/**
 * @ignore
 */

var setSingleStyle = function setSingleStyle(ele, name, value) {
  if (!ele || !ele.style) {
    throw new Error('missing HTML node element for setStyle!');
  }

  name = dashToCamel(name);
  name = checkVendorPrefix(name);

  if (typeOf(name) !== 'Undefined' && typeOf(value) !== 'Undefined') {
    ele.style[name] = tryAddPX(name, value);
  }
};
/**
 * @param {HTMLElement} ele - dom element 
 * @param {String} name - the style property name, it also can be an object to set multiple style properties
 * @param {String|Number} value - the value of the style property
 * 
 * @example
 * 
 * setStyle(document.body, 'fontSize', '16px')
 * setStyle(doucment.body, {
 *  backgroundColor: '#ffffff',
 *  minHeight: '100vh'
 * })
 */


var setStyle = function setStyle(ele, name, value) {
  if (typeOf(name) === 'Object') {
    Object.keys(name).map(function (n) {
      setSingleStyle(ele, n, name[n]);
    });
  } else {
    setSingleStyle(ele, name, value);
  }
};
/**
 * @ignore
 */

function checkVendorPrefix(name) {
  var styleKeys = getStyleKeys();
  return styleKeys[name] || name;
}
/**
 * @ignore
 */


function getStyleKeys() {
  if (styleKeys !== null) return styleKeys;

  if (typeof document === 'undefined') {
    throw new Error('missing document!');
  }

  var computedStyles = Array.from(getComputedStyle(document.body));
  var keys = Object.keys(computedStyles).filter(function (key) {
    return /^webkit(transform|transition)$/i.test(key);
  });
  styleKeys = keys.reduce(function (res, key) {
    var name = key.replace(/^webkit(\w)/, function (a, s) {
      return s.toLowerCase();
    });
    res[name] = key;
    return res;
  }, {});
  return styleKeys;
}

var properties = ['width', 'height', 'minWidth', 'maxWidth', 'minHeight', 'maxHeight', 'left', 'top', 'right', 'bottom', 'borderWidth', 'borderTopWidth', 'borderBottomWidth', 'borderLeftWidth', 'borderRightWidth', 'borderRadius', 'borderTopLeftRadius', 'borderTopRightRadius', 'borderBottomLeftRadius', 'borderBottomRightRadius', 'borderSpacing', 'fontSize', 'letterSpacing', 'margin', 'marginLeft', 'marginRight', 'marginTop', 'marginBottom', 'padding', 'paddingLeft', 'paddingRight', 'paddingTop', 'paddingBottom'];
/**
 * @ignore 
 */

function tryAddPX(name, value) {
  // check if name in the property list which can add px, if not return value 
  if (properties.indexOf(name) == -1) {
    return value;
  }

  var cValue = trim(value) - 0; // only add px to numeric value

  if (typeOf(cValue) !== 'Number') {
    return value;
  }

  return "".concat(cValue, "px");
}
/**
 * @desc set styles in batch style. It merge separate style setting into a task and execute it as soon as possible.
 * 
 * @todo 
 *  1. async set style call by delay
 *  2. merge separate call into a task
 *  3. lock the style setting when doing a task, put incoming task into a queue
 *  4. use tick as setting frame, check remaining task at every frame
 */


var TaskQueue =
/*#__PURE__*/
function (_Queue) {
  _inherits(TaskQueue, _Queue);

  function TaskQueue(options) {
    var _this;

    _classCallCheck(this, TaskQueue);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(TaskQueue).call(this, options));
    _this.waittingTasks = [];
    _this.doingTask = false;
    return _this;
  }

  _createClass(TaskQueue, [{
    key: "enqueue",
    value: function enqueue(item) {
      if (this.doingTask) {
        this.waittingTasks.push(item);
      } else {
        _get(_getPrototypeOf(TaskQueue.prototype), "enqueue", this).call(this, item);
      }

      return this;
    }
  }, {
    key: "doTask",
    value: function doTask() {
      var _this2 = this;

      if (this.doingTask) return;
      this.doingTask = true;
      var task;

      while (task = tasks.dequeue()) {
        try {
          setStyle.apply(void 0, _toConsumableArray(task));
        } catch (error) {
          console.log(error);
        }
      }

      this.doingTask = false;

      if (this.waittingTasks.length > 0) {
        this.mergeTask();
        nextFrame(function () {
          _this2.doTask();
        });
      }

      return this;
    }
  }, {
    key: "mergeTask",
    value: function mergeTask() {
      this.data = _toConsumableArray(this.data).concat(_toConsumableArray(this.waittingTasks));
      return this;
    }
  }]);

  return TaskQueue;
}(Queue);

var tasks = new TaskQueue();
var delayHandle = null;
function batchSetStyle() {
  if (delayHandle) {
    clearTimeout(delayHandle);
    delayHandle = null;
  }

  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  tasks.enqueue(args);
  delayHandle = delay(function () {
    tasks.doTask();
    delayHandle = null;
  }, 0);
}
var remSetuped = false;
function rem(_ref) {
  var _ref$designWidth = _ref.designWidth,
      designWidth = _ref$designWidth === void 0 ? 750 : _ref$designWidth,
      _ref$designDPR = _ref.designDPR,
      designDPR = _ref$designDPR === void 0 ? 1 : _ref$designDPR,
      _ref$rem2px = _ref.rem2px,
      rem2px = _ref$rem2px === void 0 ? 100 : _ref$rem2px,
      _ref$bodyFontSize = _ref.bodyFontSize,
      bodyFontSize = _ref$bodyFontSize === void 0 ? 14 : _ref$bodyFontSize,
      _ref$maxAdaptedWidth = _ref.maxAdaptedWidth,
      maxAdaptedWidth = _ref$maxAdaptedWidth === void 0 ? 1920 : _ref$maxAdaptedWidth,
      _ref$minAdaptedWidth = _ref.minAdaptedWidth,
      minAdaptedWidth = _ref$minAdaptedWidth === void 0 ? 0 : _ref$minAdaptedWidth,
      win = _ref.win,
      doc = _ref.doc;
  var documentElement = doc.documentElement;

  var dpr = function dpr() {
    return win.devicePixelRatio || 1;
  };

  var expectedWidth = designWidth * designDPR;
  var expectedDocumentElementFontSize = rem2px;
  !remSetuped && setBodyFontSize();

  function setDocumentElementFontSize() {
    var adaptedWidth = documentElement.clientWidth;
    adaptedWidth = Math.max(minAdaptedWidth, Math.min(adaptedWidth, maxAdaptedWidth));
    var adaptedDocumentElementFontSize = expectedDocumentElementFontSize * adaptedWidth / expectedWidth;
    setStyle(documentElement, 'font-size', adaptedDocumentElementFontSize);
    console.log("dpr: ".concat(dpr(), ", expectedWidth: ").concat(expectedWidth, ", deFontSize: ").concat(expectedDocumentElementFontSize, ", adaptedWidth: ").concat(adaptedWidth, ", adaptedFontSize: ").concat(adaptedDocumentElementFontSize));
  }

  function setBodyFontSize() {
    if (doc.body) {
      setStyle(doc.body, 'font-size', bodyFontSize);
    } else {
      doc.addEventListener('DOMContentLoaded', setBodyFontSize);
    }
  }

  if (!remSetuped) {
    setDocumentElementFontSize(); // reset rem unit on page resize

    window.addEventListener('resize', setDocumentElementFontSize);
    window.addEventListener('pageshow', function (e) {
      if (e.persisted) {
        setDocumentElementFontSize();
      }
    });
  }

  remSetuped = true;
}

var drawLine = function drawLine(points, end) {
  var len = points.length;
  var data = [];
  var index, point;

  for (index = 0; index < len; index++) {
    point = points[index];
    data.push("".concat(point[0], ",").concat(point[1]));
  }

  return "M".concat(data.join('L')).concat(end ? 'Z' : '');
};
var drawArc = function drawArc(_ref) {
  var innerRadius = _ref.innerRadius,
      outerRadius = _ref.outerRadius,
      startAngle = _ref.startAngle,
      endAngle = _ref.endAngle;
  var angle2Degree = Math.PI / 180;
  startAngle -= 90 * angle2Degree;
  endAngle -= 90 * angle2Degree;
  var outerArcStart = {
    x: outerRadius * Math.cos(startAngle),
    y: outerRadius * Math.sin(startAngle)
  };
  var outerArcEnd = {
    x: outerRadius * Math.cos(endAngle),
    y: outerRadius * Math.sin(endAngle)
  };
  var largeArc = (endAngle - startAngle) / angle2Degree > 180 ? 1 : 0;
  var innerArcStart, innerArcEnd;
  var path = '';

  if (innerRadius > 0) {
    innerArcStart = {
      x: innerRadius * Math.cos(startAngle),
      y: innerRadius * Math.sin(startAngle)
    };
    innerArcEnd = {
      x: innerRadius * Math.cos(endAngle),
      y: innerRadius * Math.sin(endAngle) // move to outer arc start point 

    };
    path += "M".concat(outerArcStart.x, ",").concat(outerArcStart.y); // draw a arc to outer arc end point

    path += "A".concat(outerRadius, ",").concat(outerRadius, " 0 ").concat(largeArc, ",1 ").concat(outerArcEnd.x, ",").concat(outerArcEnd.y); // draw a line to inter arc end point 

    path += "L".concat(innerArcEnd.x, ",").concat(innerArcEnd.y); // draw a arc to inter arc start point

    path += "A".concat(innerRadius, ",").concat(innerRadius, " 0 ").concat(largeArc, ",0 ").concat(innerArcStart.x, ",").concat(innerArcStart.y); // end the path

    path += 'Z';
  } else {
    // move to the origin of circle
    path += 'M0,0'; // draw a line to arc start

    path += "L".concat(outerArcStart.x, ",").concat(outerArcStart.y); // draw a arc to arc end

    path += "A".concat(outerRadius, ",").concat(outerRadius, " 0 ").concat(largeArc, ",1 ").concat(outerArcEnd.x, ",").concat(outerArcEnd.y); // end path

    path += 'Z';
  }

  return path;
};
var drawEquilateral = function drawEquilateral(_ref2) {
  var radius = _ref2.radius,
      _ref2$startAngle = _ref2.startAngle,
      startAngle = _ref2$startAngle === void 0 ? 0 : _ref2$startAngle,
      _ref2$sidesNum = _ref2.sidesNum,
      sidesNum = _ref2$sidesNum === void 0 ? 3 : _ref2$sidesNum;
  var path = '';
  var meanAngle = 2 * Math.PI / sidesNum;
  var angleFixed = deg2angle(-90);

  for (var i = 0; i < sidesNum; i++) {
    var action = void 0,
        angle = void 0,
        x = void 0,
        y = void 0;
    action = i == 0 ? 'M' : 'L';
    angle = startAngle + angleFixed + meanAngle * i;
    x = radius * Math.cos(angle);
    y = radius * Math.sin(angle);
    path += "".concat(action).concat(x, ",").concat(y);
  }

  return path + 'Z';
};

var sessStore = null;

function getSessStore() {
  if (sessStore) {
    return sessStore;
  }

  if (typeof window.sessionStorage !== 'undefined') {
    sessStore = window.sessionStorage;
  }

  return sessStore;
}

var sessionStore = {
  get: function get(key) {
    var store = getSessStore();
    if (!store) return '';
    var value = store.getItem(key) || '';
    return decodeJSON(value);
  },
  set: function set(key, value) {
    var store = getSessStore();
    if (!store) return;

    if (isArray(value) || isObject(value)) {
      value = encodeJSON(value);
    }

    store.setItem(key, value);
  },
  remove: function remove(key) {
    var store = getSessStore();
    if (!store) return;
    store.removeItem(key);
  },
  clear: function clear() {
    var store = getSessStore();
    if (!store) return;
    store.clear();
  }
};

function interpolateNumber(start, end) {
  return function (t) {
    return start + (end - start) * t;
  };
}

exports.Tween = Tween;
exports.Deceleration = Deceleration;
exports.Queue = Queue;
exports.tick = tick;
exports.padStart = _padStart;
exports.padEnd = _padEnd;
exports.typeOf = typeOf;
exports.isObject = isObject;
exports.isArray = isArray;
exports.isString = isString;
exports.isNumber = isNumber;
exports.isFunction = isFunction;
exports.isEmptyArray = isEmptyArray;
exports.isEmptyObject = isEmptyObject;
exports.isEmptyString = isEmptyString;
exports.isEmpty = isEmpty;
exports.hasValue = hasValue;
exports.isSet = isSet;
exports.isUnset = isUnset;
exports.isEmail = isEmail;
exports.isUInt = isUInt;
exports.deepCopy = deepCopy;
exports.deepEqual = deepEqual;
exports.shallowEqual = shallowEqual;
exports.assert = assert;
exports.randInt = randInt;
exports.randStr = randStr;
exports.compact = compact;
exports.addKey = addKey;
exports.head = head;
exports.tail = tail;
exports.middle = middle;
exports.midIndex = midIndex;
exports.quickSort = quickSort;
exports.nlargest = nlargest;
exports.nsmallest = nsmallest;
exports.shuffle = shuffle;
exports.uniq = uniq;
exports.intersection = intersection;
exports.difference = difference;
exports.union = union;
exports.groupBy = groupBy;
exports.bounds = bounds;
exports.contains = contains;
exports.excludeProps = excludeProps;
exports.getProp = getProp;
exports.obj2qs = obj2qs;
exports.selectProps = selectProps;
exports.propCompact = propCompact;
exports.extend = extend;
exports.echo = echo;
exports.noop = noop;
exports.delay = delay;
exports.compose = compose;
exports.debounce = debounce;
exports.guard = guard;
exports.charLength = charLength;
exports.trim = trim;
exports.dashToCamel = dashToCamel;
exports.padZero = padZero;
exports.toFixed = toFixed;
exports.addComma = addComma;
exports.angle2deg = angle2deg;
exports.deg2angle = deg2angle;
exports.currentTime = currentTime$1;
exports.timestamp = timestamp;
exports.IDFactory = IDFactory;
exports.parseQuery = parseQuery;
exports.queryJoin = queryJoin;
exports.sample1 = sample1;
exports.sample = sample;
exports.strToDate = strToDate;
exports.strToTime = strToTime;
exports.timeToStr = timeToStr;
exports.reverse = reverse;
exports.encodeJSON = encodeJSON;
exports.decodeJSON = decodeJSON;
exports.safeDivide = safeDivide;
exports.setStyle = setStyle;
exports.batchSetStyle = batchSetStyle;
exports.rem = rem;
exports.drawLine = drawLine;
exports.drawArc = drawArc;
exports.drawEquilateral = drawEquilateral;
exports.sessionStore = sessionStore;
exports.interpolateNumber = interpolateNumber;
